
def fix_scaling(feature, threshold=0.0):
    """
    对数据框中的数据进行中心化处理

    Args:
        feature: 待处理的数据框
        threshold: 低方差过滤阈值
    Returns:
        feature: 已处理数据框
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    scale_result = {"result": {}, "drop": []}
    for c in numeric_var:
        sm = feature[c].describe()
        scale_result[c] = {'mean': sm['mean'], 'std': sm['std']}
        feature[c] = feature[c].apply(lambda x: (x - sm['mean'])/sm['std'] if x else x)
        if sm['std']**2 < threshold:
            scale_result['drop'].append(c)
    return feature, scale_result



from math import log
import pandas as pd
import pickle

def make_score(odds, score, pdo):
    # odds = P_0 / P_1
    b = float(pdo) / log(2)
    a = float(score) - b * log(float(odds))
    return a, b


def make_card(coefs, woe_result, odds, score, pdo):
    # odds = P_0 / P_1
    a, b = make_score(odds, score, pdo)
    score_card = pd.DataFrame([['base_score', '-', '-', int(a - b * coefs['intercept_'])]],
                              columns=['Variables', 'Bins', 'Woe', 'Score'])
    for v in list(coefs.keys())[:-1]:
        if v in woe_result.keys():
            woe_result[v].insert(loc=0, column='Variables', value=v)
            woe_result[v].rename(columns={v: "Bins", 'woe_i': 'Woe'}, inplace=True)
            woe_result[v]['Score'] = woe_result[v]['Woe'].apply(lambda x: int(-x * b * coefs[v]))
            score_card = pd.concat([score_card, woe_result[v]])
        else:
            df = pd.DataFrame([[v, '-', '-', int(- b * coefs[v])]], columns=['Variables', 'Bins', 'Woe', 'Score'])
            score_card = pd.concat([score_card, df])
    score_card.to_excel('result/card.xlsx', index=False)
    with open("result/card.pickle", 'wb') as f:
        f.write(pickle.dumps(score_card))
    return score_card



from statsmodels.iolib.smpickle import load_pickle
import pandas as pd
import mksc
from mksc.model import training
from mksc.feature_engineering import preprocess
from mksc.feature_engineering import transform
from custom import Custom

def main(**kwargs):

    feature_engineering = load_pickle('result/feature_engineering.pickle')
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]

    # 自定义数据清洗
    feature, label = Custom.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # 自定义特征组合模块
    feature = Custom.feature_combination(feature)
    feature.drop(datetime_var,  axis=1, inplace=True)

    # 数据处理
    feature = transform(feature, feature_engineering)

    # 模型训练
    training(feature, label, **kwargs)


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 2:
        main(model_name=sys.argv[1])
    else:
        main()



import numpy as np
from mksc.feature_engineering import binning
from math import log
import pandas as pd

def transform(feature, feature_engineering):

    # One-Hot编码
    category_var = feature.select_dtypes(include=['object']).columns
    feature[category_var].fillna("NA", inplace=True)
    if not feature[category_var].empty:
        feature = pd.concat([feature, pd.get_dummies(feature[category_var])], axis=1)
    feature.drop(category_var, axis=1, inplace=True)
    feature = feature[feature_engineering['feature_selected']]

    # 极端值处理
    abnormal_value = feature_engineering['abnormal_value']
    for c in set(feature.columns) & set(abnormal_value['replace']):
        max_ = abnormal_value['result'][c]['max']
        min_ = abnormal_value['result'][c]['min']
        feature.loc[:, c] = feature.loc[:, c].apply(lambda x: x if (x < max_) & (x > min_) else np.nan)

    # 缺失值处理
    missing_filling = feature_engineering['missing_filling']
    for c in set(feature.columns) & set(missing_filling['replace']):
        fill_number = abnormal_value['result'][c]['fill_number']
        feature[c].fillna(fill_number, inplace=True)

    # 归一化处理
    scale_result = feature_engineering['scale_result']['result']
    for c in set(feature.columns) & set(scale_result.keys()):
        mean = scale_result[c]['mean']
        std = scale_result[c]['std']
        feature.loc[:, c] = feature.loc[:, c].apply(lambda x: (x - mean)/std if x else x)

    # 正态化处理
    standard_lambda = feature_engineering['standard_lambda']
    for c in set(feature.columns) & set(standard_lambda.keys()):
        _lambda = standard_lambda[c]
        feature.loc[:, c] = feature.loc[:, c] + 0.5
        feature.loc[:, c] = feature.loc[:, c].apply(lambda x: (x**_lambda - 1) / _lambda if _lambda > 0 else log(x))

    # woe转化
    woe_result = feature_engineering['woe_result']
    bin_result = feature_engineering['bin_result']
    feature = binning.woe_transform(feature, woe_result, bin_result)

    return feature



from math import log
from copy import deepcopy
import pandas as pd
from sklearn.tree import DecisionTreeClassifier

def tree_binning(label, feature):
    """
    通过决策树执行模型分箱函数
     y_i: 第i组中1的个数
     y_T: 所有1的个数
     n_i: 第i组中0的个数
     n_T: 所有0的个数
     WOE_i = ln(P_yi/P_ni) = ln(y_i*n_T/y_T/n_i)
     IV = sum(IV_i) = sum((P_yi-P_ni)*WOE_i) = sum((y_i/y_T-n_i/n_T)*WOE_i)

    Args:
        label: 待分箱特征的标签
        feature: 待分箱特征的数据框

    Returns:
        bin_result: 分箱结果数据
        iv_result: IV值统计结果
        woe_result: WOE值统计结果
        woe_adjust_result: 需要调整分箱的结果
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    label_name = label.name
    y_t = label[label.values == 1].count()
    n_t = label[label.values == 0].count()
    bin_result = {'result': {}, 'error': []}
    iv_result = {}
    woe_result = {}
    for c in numeric_var:
        temp = pd.concat([feature[c], label], axis=1)
        # 统计空值分布
        binning_data_na = temp[temp[c].isna()]
        if len(binning_data_na):
            y_na = binning_data_na[label_name][binning_data_na[label_name].values == 1].count()
            n_na = binning_data_na[label_name][binning_data_na[label_name].values == 0].count()
            if y_na and n_na:
                df_na = pd.DataFrame([['nan', y_na, n_na]], columns=[c, 'y_i', 'n_i'])
            else:
                bin_result['error'].append(c)
                continue
        else:
            df_na = pd.DataFrame(columns=[c, 'y_i', 'n_i'])

        # cart树确定分箱边界
        binning_data = temp.dropna(subset=[c])
        model = DecisionTreeClassifier(min_samples_leaf=0.05, min_samples_split=0.02, max_depth=5, max_leaf_nodes=15)
        model.fit(binning_data[c].values.reshape(-1, 1), binning_data[label_name].values)
        boundary = model.tree_.threshold
        boundary = sorted(list(boundary[boundary != -2]))
        boundary = [float('-inf')] + boundary + [float('inf')]
        bin_result['result'][c] = boundary

        # 统计分箱结果
        df = pd.merge(pd.cut(temp[c], boundary), temp[label_name], right_index=True, left_index=True)
        df['count'] = 1
        df = df.groupby([c, label_name]).count().unstack(level=1)
        df.reset_index(inplace=True)
        df.columns = [x[0] + str(x[1]) for x in df.columns.ravel()]
        df.rename(columns={'count0': 'n_i', 'count1': 'y_i'}, inplace=True)
        df = pd.concat([df_na, df])
        df.reset_index(inplace=True)
        df['woe_i'] = df.apply(lambda x: log(x['y_i']) + log(n_t) - log(y_t) - log(x['n_i']), axis=1)
        df['iv_i'] = df.apply(lambda x: (x['y_i']/y_t-x['n_i']/n_t)*x['woe_i'], axis=1)
        woe_result[c] = df[[c, 'woe_i']]
        iv_result[c] = df['iv_i'].sum()

    # 分箱调整：筛选出分箱中存在一类的特征，转化为规则预测
    woe_adjust_result = []
    for w in deepcopy(woe_result):
        if woe_result[w].woe_i.isna().any():
            woe_adjust_result.append(w)
            woe_result.pop(w)

    return bin_result, iv_result, woe_result, woe_adjust_result



import numpy as np

def fix_abnormal_value(feature, threshold=0.05):
    """
    修正数据框中的数值型变量中的异常值

    Args:
        feature: 待修正的数据框
        threshold: 异常值替换阈值

    Returns:
        feature: 已处理数据框
        abnormal_value: 异常值统计结果
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    abnormal_value = {'result': {}, 'replace': []}
    for c in numeric_var:
        sm = feature[c].describe()
        iqr = sm['75%'] - sm['25%']
        min_ = sm['25%'] - 1.5*iqr
        max_ = sm['75%'] + 1.5*iqr
        abnormal_value_indexes = list(feature.loc[(feature[c] <= min_) | (feature[c] >= max_)].index)
        abnormal_value_length = len(abnormal_value_indexes)
        abnormal_value_rate = abnormal_value_length/feature[c].count()
        abnormal_value['result'][c] = {'abnormal_value_list': abnormal_value_indexes,
                                       'abnormal_value_length': abnormal_value_length,
                                       'abnormal_value_rate': abnormal_value_rate,
                                       'max': max_,
                                       'min': min_}
        if abnormal_value_rate <= threshold:
            abnormal_value['replace'].append(c)
            feature.loc[:, c] = feature.loc[:, c].apply(lambda x: x if (x < max_) & (x > min_) else np.nan)
    return feature, abnormal_value



from statsmodels.iolib.smpickle import load_pickle
import pandas as pd
import configparser
from datetime import date
from math import log
import os
import mksc
from mksc.feature_engineering import preprocess
from mksc.feature_engineering import scoring
from mksc.feature_engineering import transform
from custom import Custom

def main(model_name, card=False, score=True):
    # 数据、模型加载
    model, threshold = load_pickle(f'result/{model_name}.pickle')

    feature_engineering = load_pickle('result/feature_engineering.pickle')
    data = mksc.load_data("apply")
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = []

    # 自定义数据清洗
    feature, label = Custom.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # 自定义特征组合模块
    feature = Custom.feature_combination(feature)
    feature.drop(datetime_var, axis=1, inplace=True)

    # 数据处理
    feature = transform(feature, feature_engineering)

    # 应用预测
    res_label = pd.DataFrame(model.predict(feature), columns=['label_predict'])
    res_prob = pd.DataFrame(model.predict_proba(feature), columns=['probability_0', "probability_1"])
    res_prob['res_odds'] = res_prob['probability_0'] / res_prob["probability_1"]
    res_prob['label_threshold'] = res_prob['probability_1'].apply(lambda x: 0 if x < threshold else 1)
    res = pd.concat([data, feature, res_label, res_prob], axis=1)

    if card and model_name == 'lr':
        # 转化评分
        score_card = load_pickle('result/card.pickle')
        res = scoring.transform_score(res, score_card)

    if score:
        cfg = configparser.ConfigParser()
        cfg.read(os.path.join(os.getcwd(), 'config', 'configuration.ini'), encoding='utf_8_sig')
        odds = cfg.get('SCORECARD', 'odds')
        score = cfg.get('SCORECARD', 'score')
        pdo = cfg.get('SCORECARD', 'pdo')
        a, b = scoring.make_score(odds, score, pdo)
        res['Score'] = res_prob['res_odds'].apply(lambda x: a + b * log(float(x)))

    # 结果保存
    res['load_date'] = str(date.today())
    mksc.save_result(res, remote=True)
    mksc.save_result(res, filename="apply_result.csv")


if __name__ == '__main__':
    import sys
    if len(sys.argv) == 2:
        main(sys.argv[1])



import pandas as pd
import pandas_profiling as pp
from statsmodels.iolib.smpickle import save_pickle
import mksc

def main(report=True, apply=True):
    """
    探索性数据分析主程序入口
    Args:
        report: 是否保留分析报告，由于该报告对大数据集会占用大量时间，默认生成
        apply: 是否读取应用集并保存pickle文件
    生成以下四份文件：
        1、数据对象文件： “data/data.pickle”
        2、变量类型配置文件： “config/variable_type.csv”
        3、样例数据： “data/sample.xlsx”
        4、结果报告 ： “result/report.html”
    """
    # 加载数据并保存本地
    data = mksc.load_data(mode="all", local=False)
    save_pickle(data, 'data/data.pickle')

    # 生成变量类别配置文件
    # 变量是否保留进行特征工程(isSave): 0-不保留；1-保留
    # 变量类型(Type): numeric-数值类型；category-类别类型；datetime-日期类型；label-标签列
    res = pd.DataFrame(zip(data.columns,
                           [1]*len(data.columns),
                           ['category']*len(data.columns),
                           ['']*len(data.columns)),
                       columns=['Variable', 'isSave:[0/1]', 'Type:[numeric/category/datetime/label]', 'Comment'])
    res.to_csv("config/variable_type.csv", index=False)

    # 抽样探索
    sample = data.sample(min(len(data), 1000))
    sample.reset_index(drop=True, inplace=True)
    with pd.ExcelWriter('data/sample.xlsx') as writer:
        sample.to_excel(writer, sheet_name="抽样数据", index=False)
        sample.corr().to_excel(writer, sheet_name="相关系数", index=False)
        (sample
            .select_dtypes(exclude=['object', 'datetime'])
            .describe()
            .to_excel(writer, sheet_name="数值数据汇总"))
        (sample
            .select_dtypes(include=['object', 'datetime'])
            .describe()
            .to_excel(writer, sheet_name="分类数据汇总"))

    # 保存分析报告
    if report and (not sample.empty):
        report = pp.ProfileReport(sample)
        report.to_file('result/report.html')

    # 保存应用集，注意规模
    if apply:
        apply = mksc.load_data(mode="apply", local=False)
        save_pickle(apply, 'data/apply.pickle')


if __name__ == '__main__':
    main(report=False, apply=False)



import pickle
import pandas as pd
from mksc.feature_engineering import seletction
from mksc.feature_engineering import values
from mksc.feature_engineering import binning


class FeatureEngineering(object):

    def __init__(self, feature, label, missing_threshold=(0.9, 0.05), distinct_threshold=0.9, unique_threshold=0.9,
                 abnormal_threshold=0.05, correlation_threshold=0.7, variance_threshold=0.05):
        self.feature = feature
        self.label = label
        self.missing_threshold = missing_threshold
        self.distinct_threshold = distinct_threshold
        self.unique_threshold = unique_threshold
        self.abnormal_threshold = abnormal_threshold
        self.correlation_threshold = correlation_threshold
        self.variance_threshold = variance_threshold
        self.threshold = {"missing_threshold": self.missing_threshold,
                          "distinct_threshold": self.distinct_threshold,
                          "unique_threshold": self.unique_threshold,
                          "abnormal_threshold": self.abnormal_threshold,
                          "correlation_threshold": self.correlation_threshold}

    def run(self):
        """
        特征工程过程函数,阈值参数可以自定义修改
        1. 特征组合
        2. 基于统计特性特征选择：缺失率、唯一率、众数比例
        3. 缺失值处理
        TODO 异常值处理
        4. 极端值处理
        5. 正态化处理
        6. 归一化处理
        7. 最优分箱
        8. IV筛选
        TODO PSI筛选
        9. 相关性筛选
        10. woe转化
        11. One-Hot
        12. TODO 降维

        Returns:
            feature: 已完成特征工程的数据框
            label: 已完成特征工程的标签列
        """
        feature = self.feature
        label = self.label

        # 基于缺失率、唯一率、众数比例统计特征筛选
        missing_value = seletction.get_missing_value(feature, self.missing_threshold[0])
        distinct_value = seletction.get_distinct_value(feature, self.distinct_threshold)
        unique_value = seletction.get_unique_value(feature, self.unique_threshold)
        feature.drop(set(missing_value['drop'] + distinct_value['drop'] + unique_value['drop']), axis=1, inplace=True)

        # One-Hot编码
        category_var = feature.select_dtypes(include=['object']).columns
        if not category_var.empty:
            feature[category_var].fillna("NA", inplace=True)
            feature = pd.concat([feature, pd.get_dummies(feature[category_var])], axis=1)
            feature.drop(category_var, axis=1, inplace=True)

        # 缺失值处理
        feature, missing_filling = values.fix_missing_value(feature, self.missing_threshold[1])

        # 极端值处理
        feature, abnormal_value = values.fix_abnormal_value(feature, self.abnormal_threshold)

        # 标准化处理
        feature, scale_result = values.fix_scaling(feature, self.variance_threshold)
        feature.drop(scale_result['drop'], axis=1, inplace=True)

        # 正态化处理
        feature, standard_lambda = values.fix_standard(feature)

        # 数值特征最优分箱，未处理的变量，暂时退出模型
        bin_result, iv_result, woe_result, woe_adjust_result = binning.tree_binning(label, feature)
        bin_error_drop = bin_result['error'] + woe_adjust_result

        # IV筛选
        iv_drop = list(filter(lambda x: iv_result[x] < 0.02, iv_result))
        feature.drop(iv_drop + bin_error_drop, inplace=True, axis=1)

        # 相关性筛选
        cor_drop = seletction.get_cor_drop(feature, iv_result, self.correlation_threshold)
        feature.drop(cor_drop, inplace=True, axis=1)

        # woe转化
        feature = binning.woe_transform(feature, woe_result, bin_result)

        # 降维
        if feature.shape[1] > 300:
            pass
            # 重采样
            # # 逐步回归筛选
            # feature_selected = seletction.stepwise_selection(feature, label)
        feature_selected = feature.columns

        # 中间结果保存
        result = {"missing_value": missing_value,
                  "distinct_value": distinct_value,
                  "unique_value": unique_value,
                  "abnormal_value": abnormal_value,
                  "missing_filling": missing_filling,
                  'standard_lambda': standard_lambda,
                  'scale_result': scale_result,
                  "bin_result": bin_result,
                  "iv_result": iv_result,
                  "woe_result": woe_result,
                  "woe_adjust_result": woe_adjust_result,
                  "bin_error_drop": bin_error_drop,
                  "iv_drop": iv_drop,
                  "cor_drop": cor_drop,
                  "feature_selected": feature_selected
                  }
        with open('result/feature_engineering.pickle', 'wb') as f:
            f.write(pickle.dumps(result))

        with open('result/apply_sql.txt', 'w') as f:
            sql = ",".join(result["feature_selected"])
            f.write(f"select {sql} from ")



from statsmodels.iolib.smpickle import load_pickle
from mksc.feature_engineering import scoring
import configparser
import os

def main():
    """
    评分卡制作主程序入口
    """
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(os.getcwd(), 'config', 'configuration.ini'), encoding='utf_8_sig')
    odds = cfg.get('SCORECARD', 'ODDS')
    score = cfg.get('SCORECARD', 'SCORE')
    pdo = cfg.get('SCORECARD', 'PDO')
    feature_engineering = load_pickle("result/feature_engineering.pickle")
    woe_result = feature_engineering["woe_result"]
    model = load_pickle("result/lr.pickle")
    coefficient = list(zip(feature_engineering["feature_selected"], list(model.coef_[0])))
    coefficient.append(("intercept_", model.intercept_[0]))
    coefs = dict(coefficient)
    scoring.make_card(coefs, woe_result, odds, score, pdo)


if __name__ == "__main__":
    main()



import sys
import os
import shutil

def make_workspace(name):
    """
    创建项目的工作目录与脚本文件

    Args:
        name: 项目名
    """

    if os.path.exists(name):
        raise TypeError(f"Folder [{name}] is exists already, Please check out the work path")
    else:
        templates = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'template')
        project = os.path.join(os.getcwd(), name)
        shutil.copytree(templates, project)

    os.mkdir(os.path.join(os.getcwd(), name, 'result'))
    os.mkdir(os.path.join(os.getcwd(), name, 'log'))

    with open(os.path.join(project, 'config', 'configuration.ini'), "r", encoding='utf-8') as f:
        content = f.read()
    with open(os.path.join(project, 'config', 'configuration.ini'), "w", encoding='utf-8') as f:
        content = content % (os.path.join(os.getcwd(), name))
        f.write(content)

def main():
    """
    命令行工具程序主入口
    """
    if len(sys.argv) == 1:
        return "CMD FORMAT: \n\tmksc project_name1 project_name2 ...\nPlease delivery one argument at least"
    else:
        for project_name in sys.argv[1:]:
            make_workspace(project_name)


if __name__ == "__main__":
    main()



from mksc.core import reader
import os

def save_result(data, filename=None, remote=False):
    cfg = reader.config()
    if remote:
        data.to_sql(cfg.get('DATABASE', 'SCHEMA_NAME'), cfg.get('DATABASE', 'SAVE_ENGINE_URL'), if_exists='replace')
    else:
        data.to_csv(os.path.join(cfg.get('PATH', 'WORK_DIR'), 'result', filename), index=False)



from statsmodels.iolib.smpickle import load_pickle
import pandas as pd
import os
import configparser

def file(filename):
    if '.csv' in filename:
        data = pd.read_csv(filename)
        return data
    elif '.pickle' in filename:
        data = load_pickle(filename)
        return data
    elif ('.xls' in filename) or ('.xlsx' in filename):
        data = pd.read_excel(filename)
        return data
    else:
        print("Wrong Data Type")

def config():
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(os.getcwd(), 'config', 'configuration.ini'), encoding='utf_8_sig')
    return cfg



import pandas as pd
import statsmodels.api as sm


def stepwise_selection(x, y, threshold_in=0.01, threshold_out=0.05):
    """
    通过逐步回归筛选特征变量

    Args:
        x: 待筛选的特征数据框
        y: 数据框标签
        threshold_in: 进入模型的最大P值
        threshold_out: 退出模型的最小P值

    Returns:
        included：特征名称列表
    """
    included = []
    numeric_var = x.select_dtypes(exclude=['object', 'datetime']).columns
    while True:
        changed = False
        # forward step
        excluded = list(set(numeric_var)-set(included))
        new_pval = pd.Series(index=excluded)
        for new_column in excluded:
            model = sm.Logit(y, sm.add_constant(pd.DataFrame(x[included+[new_column]]))).fit()
            new_pval[new_column] = model.pvalues[new_column]
        best_pval = new_pval.min()
        if best_pval < threshold_in:
            best_feature = new_pval.idxmin()
            included.append(best_feature)
            changed = True
 
        # backward step
        model = sm.Logit(y, sm.add_constant(pd.DataFrame(x[included]))).fit()
        # use all coefs except intercept
        p_values = model.pvalues.iloc[1:]
        # null if p-values is empty
        worst_pval = p_values.max()
        if worst_pval > threshold_out:
            changed = True
            worst_feature = p_values.idxmax()
            included.remove(worst_feature)
        if not changed:
            break
    return included



import eda
import train
import score
import apply
import feature

def main():
    tag = input(">>> step 1: 请完成项目配置configuration.ini，完成后请键入【Y/y】继续，任意键退出")
    if tag.upper() != "Y":
        return SystemExit(">>> 请完成项目配置")

    tag = input(">>> step 2: EDA过程跳过请输入【Y/y】")
    if tag.upper() != "Y":
        report = input(">>> 报告生成将花费大量时间，回车确认取消报告，任意键回车继续生成")
        apply = input(">>> 应用集规模过大将无法一次读取保存，回车确认取消读取保存，任意键回车继续生成")
        eda.main(report, apply)

    tag = input(">>> step 3: 请完成特征配置variable_type.csv以及自定义配置custom.py，完成请输入【Y/y】，跳过特征过程请输入【S/s】")
    if tag.upper() == "Y":
        print(">>> 特征过程...")
        feature.main()
    elif tag.upper() != "S":
        return SystemExit(">>> 错误参数，退出程序")

    tag = input(">>> step 4: 训练过程，跳过过程请输入【S/s】")
    if tag.upper() != "S":
        print(">>> 训练过程...")
        train.main()
    elif tag.upper() != "S":
        return SystemExit(">>> 错误参数，退出程序")

    tag = input(">>> step4: 打分过程，跳过请输入【Y/y】")
    if tag.upper() != "Y":
        score.main()

    tag = input(">>> step6: 应用过程，跳过请输入【Y/Y】")
    if tag.upper() != "Y":
        apply.main()


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import re

class Custom(object):
    """
    自定义预处理类函数封装
    """
    def __init__(self):
        pass

    @staticmethod
    def clean_data(feature, label):
        """
        基于单元格与行操作的数据清洗

        Args:
            feature: 待清洗特征数据框
            label: 待清洗标签序列

        Returns:
            feature_tmp: 已清洗特征数据框
            label_tmp: 已清洗标签序列
        """
        feature_tmp = feature.copy()
        label_tmp = label.copy()
        # 默认值处理
        # 单位处理
        # 正则处理
        # 业务缺失值补充
        # 业务异常值剔除
        return feature_tmp, label_tmp

    @staticmethod
    def feature_combination(feature):
        """
        基于列操作的数据清洗与特征构造

        Args:
            feature: 待清洗特征数据框

        Returns:
            feature_tmp: 已清洗特征数据框
        """
        feature_tmp = feature.copy()
        # 构造衍生变量
        # 日期变量处理，此处无需丢弃变量
        date_var = ['day', 'dayofweek', 'dayofyear', 'days_in_month', 'is_leap_year', 'is_month_end',
                    'is_month_start', 'is_quarter_end', 'is_quarter_start', 'is_year_end',
                    'is_year_start', 'month', 'quarter', 'week', 'weekday', 'weekofyear', 'year']
        return feature_tmp

class CustomModel(object):
    """
    自定义模型, 必须包括以下三个方法
    """
    def fit(self, x, y):
        pass

    def predict(self, x):
        return

    def predict_proba(self, x):
        return

class CustomApply(object):
    """
    自定义应用集处理
    """
    def load_data(self):
        pass

    def load_model(self):
        pass

    def predict(self):
        pass


if __name__ == "__main__":

    import mksc
    from mksc.feature_engineering import preprocess
    # 加载数据、变量类型划分、特征集与标签列划分
    data = mksc.load_data()
    numeric_var, category_var, datetime_var, label_var = preprocess.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')



from .statistics import (
                        get_cor_drop,
                        get_distinct_value,
                        get_missing_value,
                        get_unique_value
)

from .model import stepwise_selection



import random

def get_missing_value(feature, threshold=0.9):
    """
    返回缺失值统计信息

    Args:
        feature: 待处理特征数据框
        threshold: 缺失值阈值

    Returns:
        missing_value：缺失值统计信息
    """
    missing_value = {'drop': [], 'result': {}}
    for c in feature:
        missing_rate = round(feature[c].isna().sum()/len(feature), 3)
        missing_value['result'][c] = missing_rate
        if missing_rate >= threshold:
            missing_value['drop'].append(c)
    missing_value['drop_number'] = len(missing_value['drop'])
    return missing_value
    
def get_distinct_value(feature, threshold=0.9):
    """
    返回唯一率统计信息

    Args:
        feature: 待处理特征数据框
        threshold: 唯一率阈值

    Returns:
        missing_value：唯一率统计信息
    """
    distinct_value = {'drop': [], 'result': {}}
    for c in feature:
        distinct_rate = round(len(feature[c].drop_duplicates())/feature[c].count(), 3)
        distinct_value['result'][c] = distinct_rate
        if distinct_rate > threshold:
            distinct_value['drop'].append(c)
    distinct_value['drop_number'] = len(distinct_value['drop'])
    return distinct_value

def get_unique_value(feature, threshold=0.9):
    """
    返回众数比率统计信息

    Args:
        feature: 待处理特征数据框
        threshold: 众数比率阈值

    Returns:
        missing_value：众数比率统计信息
    """
    unique_value = {'drop': [], 'result': {}}
    for c in feature:
        sm = feature[c].astype('O').describe()
        unique_rate = round(sm['freq']/sm['count'], 3)
        unique_value['result'][c] = unique_rate
        if unique_rate > threshold:
            unique_value['drop'].append(c)
    unique_value['drop_number'] = len(unique_value['drop'])
    return unique_value

def iv_compare(iv_result, var1, var2):
    """
    返回特征var1, var2中IV较大的变量名
    Args:
        iv_result: IV值统计信息
        var1: 变量名1
        var2: 变量名2

    Returns:
        IV较大的变量名
    """
    if var1 in iv_result.keys() and var2 in iv_result.keys():
        var1_iv = iv_result[var1]
        var2_iv = iv_result[var2]
        if var1_iv > var2_iv:
            return var1
        else:
            return var2
    elif var1 not in iv_result.keys() and var2 not in iv_result.keys():
        return random.choice([var1, var2])
    else:
        return None

def get_cor_drop(feature, iv_result, threshold=0.7):
    """
    返回相关性又高，IV值相对也高德变量列表
    Args:
        feature: 待处理数据框
        iv_result: IV值信息
        threshold: 相关系数缺失阈值
    Returns:
        cor_drop: 相关性与IV值相对较高的变量列表
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    cor = feature[numeric_var].corr()
    cor = cor[(abs(cor) > threshold) & (cor != 1)].dropna(how='all', axis=1).dropna(how='all')
    if not cor.empty:
        cor = cor.stack().reset_index()
        cor.columns = ['Var_1', 'Var_2', 'cor']
        cor = cor[cor.index % 2 == 0]
        cor = cor.sort_values(by='cor', ascending=False)
        cor['smaller_iv'] = cor.apply(lambda x: iv_compare(iv_result, x['Var_1'], x['Var_2']), axis=1)
        cor_drop = list(filter(lambda x: x, set(cor['smaller_iv'].values)))
        return cor_drop
    else:
        return []



from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from xgboost.sklearn import XGBClassifier
from lightgbm import LGBMClassifier

model_mapper = {"lr": LogisticRegression,
                "dt": DecisionTreeClassifier,
                "svm": SVC,
                "rf": RandomForestClassifier,
                "gbdt": GradientBoostingClassifier,
                "xgb": XGBClassifier,
                "lgb": LGBMClassifier}

def model_choose(model_name, **kwargs):
    global model_mapper
    return model_mapper[model_name](**kwargs)



import pickle
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import accuracy_score, recall_score, auc, roc_curve
from sklearn.model_selection import cross_validate
from sklearn.model_selection import train_test_split
from imblearn.over_sampling import SMOTE
from mksc.model import model_mapper, model_choose


def training(feature, label, model_name=None, **kwargs):

    print(f"样本比例：label.sum()/len(label) ")
    # 重采样
    if "resample_threshold" in kwargs.keys():
        if (label.sum()/len(label) < kwargs["resample_threshold"]
                or label.sum()/len(label) > 1 - kwargs["resample_threshold"]):
            feature, label = SMOTE().fit_sample(feature, label)
    # 数据集划分
    x_train, x_test, y_train, y_test = train_test_split(feature, label, test_size=0.2, random_state=0)

    # 模型训练
    result = {}
    if model_name:
        model = model_choose(model_name, **kwargs)
        scores = cross_validate(model, x_train, y_train, cv=5, scoring='roc_auc')
        result[model_name] = scores['test_score'].mean()
        print(f"Model-{model_name}: roc_auc: {result[model_name]:.2f}")
        model.fit(x_train, y_train)
    else:
        for m in model_mapper:
            model = model_choose(m)
            if m == 'svm':
                model = model_choose(m, probability=True)
            scores = cross_validate(model, x_train, y_train, cv=5, scoring='roc_auc')
            result[m] = scores['test_score'].mean()
            print(f"Model-{m}: roc_auc: {result[m]:.2f}")
        model_name = max(result, key=result.get)
        model = model_choose(model_name)
        model.fit(x_train, y_train)

    # ROC曲线&阈值计算
    predict_train = np.array([i[1] for i in model.predict_proba(x_train)])
    fpr, tpr, thresholds = roc_curve(y_train.values, predict_train, pos_label=1)
    auc_score = auc(fpr, tpr)
    w = tpr - fpr
    ks_score = w.max()
    ks_x = fpr[w.argmax()]
    ks_y = tpr[w.argmax()]
    fig, ax = plt.subplots()
    ax.plot(fpr, tpr, label='AUC=%.5f' % auc_score)
    ax.set_title('Receiver Operating Characteristic')
    ax.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6))
    ax.plot([ks_x, ks_x], [ks_x, ks_y], '--', color='red')
    ax.text(ks_x, (ks_x + ks_y) / 2, '  KS=%.5f' % ks_score)
    ax.legend()
    fig.savefig(f'result/{model_name}_roc.png')

    # 模型评估
    predict_train = [0 if i[1] < thresholds[w.argmax()] else 1 for i in model.predict_proba(x_train)]
    predict_test = [0 if i[1] < thresholds[w.argmax()] else 1 for i in model.predict_proba(x_test)]

    acu_train = accuracy_score(y_train, predict_train)
    acu_test = accuracy_score(y_test, predict_test)

    sen_train = recall_score(y_train, predict_train, pos_label=1)
    sen_test = recall_score(y_test, predict_test, pos_label=1)

    spe_train = recall_score(y_train, predict_train, pos_label=0)
    spe_test = recall_score(y_test, predict_test, pos_label=0)
    print(f"K-S 值： {ks_score}")
    print(f'模型准确率：训练 {acu_train * 100:.2f}%	测试 {acu_test * 100:.2f}%')
    print(f'正例覆盖率：训练 {sen_train * 100:.2f}%	测试 {sen_test * 100:.2f}%')
    print(f'负例覆盖率：训练 {spe_train * 100:.2f}%	测试 {spe_test * 100:.2f}%')

    # 模型保存
    result = (model, thresholds[w.argmax()])
    with open(f'result/{model_name}.pickle', 'wb') as f:
        f.write(pickle.dumps(result))
